function animate(x, t, movie, desc)
%
% ANIMATE2 Visualize the walker walking down a slope. May need to edit
% depending on the model.
%
% INPUT:
% + xcycle = vector of the states at each integration step. See walk.
% + movie = the movie recording option. If movie == '', no recording is
%   made. Otherwise, this string is used as the movie's file name.
%
% Notes:
% Avoid running anything computationally intensive when you use this
% animation.
%
% Written by Eric D.B. Wendel and Robert D. Gregg and Mark Yeatman

    global L la l1 l2 lf 
    global Step_time_sum stepnum slope MovingCOP

    if nargin < 3
        movie = '';
    end
    if nargin < 4
        desc = '';
    end

    heelx = x(:,1);
    heely = x(:,2);
    heeltheta = x(:,3);
    sh = x(:,4); 
    th = x(:,5); 
    nsth = x(:,6);
    nssh = x(:,7);
    nsf = x(:,8);
    sh = sh + heeltheta;
    th = th + sh;
    nsth = nsth + th;
    nssh = nssh + nsth;
    nsf = nsf + nssh;

    % Create figure
    screenSize = get(0, 'ScreenSize');
    width = screenSize(3)/2.5;%1.5;%width = screenSize(3);
    left = 0;
    bottom = screenSize(4)/4;
    height = screenSize(4)/3;%height = screenSize(4)/1.5;
    fig=figure('Position', [left bottom width height], 'Color','w');
    set(fig,'DoubleBuffer','on');
    set(gca,'xlim',[-80 80],'ylim',[-80 80],'NextPlot','replace','Visible','off')
    refPoint = [0 0];
    camera_shift = [0,0];
    % Initialize the movie
    FramesPerSec=30;  
    nextFrameTime = 0;
    timescale=1/5;

    if ~isempty(movie)
        mov=VideoWriter(movie);
        mov.FrameRate=FramesPerSec;
        mov.Quality=100;
        open(mov)
        if ~isempty(desc)
            groupTitle = title([desc ', ' date]);
            set(groupTitle);
        end
    end

    steps=0;
    axis([0,1,0,1]);
    N=1;
    for i=1:length(t)

        % Switch stance, nonstance legs
        if i >= 2 && abs(th(i)-nsth(i-1))<=0.0002 && abs(sh(i)-nssh(i-1))<=0.0002
            steps = steps+1;
            refPoint = nsPoint+(refPoint-heelPoint);
        end

        % Determine the hip and feet positions in 2D space
        heelPoint = refPoint + [heelx(i) heely(i)];
        toePoint= heelPoint + [-lf*cos(heeltheta(i)) -lf*sin(heeltheta(i))];
        ankPoint = heelPoint + [la*sin(-heeltheta(i)) la*cos(heeltheta(i))];
        kneePoint = ankPoint + [l1*sin(-sh(i)) l1*cos(sh(i))];
        hipPoint = kneePoint + [l2*sin(-th(i)) l2*cos(th(i))];
        nskneePoint = hipPoint - [l2*sin(-nsth(i)) l2*cos(nsth(i))];
        nsankPoint = nskneePoint - [l1*sin(-nssh(i)) l1*cos(nssh(i))];
        nsPoint = nsankPoint - [la*sin(-nsf(i)) la*cos(nsf(i))];
        nstoePoint =  nsPoint-[lf*cos(-nsf(i)) lf*sin(nsf(i))];

        x_pros = [toePoint(1) heelPoint(1) ankPoint(1) kneePoint(1) hipPoint(1)];
        y_pros = [toePoint(2) heelPoint(2) ankPoint(2) kneePoint(2) hipPoint(2)];
        x_bod = [hipPoint(1) nskneePoint(1) nsankPoint(1) nsPoint(1) nstoePoint(1)];
        y_bod = [hipPoint(2) nskneePoint(2) nsankPoint(2) nsPoint(2) nstoePoint(2)];

        if (t(i) >= nextFrameTime) || i == length(t)
            nextFrameTime = nextFrameTime + timescale/FramesPerSec;
            hold off

            %Maintain leg swap in linestyle, draw links
            if ~mod(steps,2)
                plot(x_pros, y_pros, 'Color',[0.3 0.3 0.3],'LineWidth',3)
                hold on, plot(x_bod, y_bod,'k--', 'LineWidth',3)
            else
                plot(x_pros, y_pros,'k--','LineWidth',3)
                hold on, plot(x_bod, y_bod, 'Color',[0.3 0.3 0.3], 'LineWidth',3)
            end

            % Draw joint points
            plot(ankPoint(1), ankPoint(2), 'o','LineWidth',1,'MarkerSize',6,'MarkerEdgeColor','k','MarkerFaceColor','y')
            plot(kneePoint(1), kneePoint(2), 'o','LineWidth',1.5,'MarkerSize',8,'MarkerEdgeColor','k','MarkerFaceColor','y')
            plot(hipPoint(1), hipPoint(2), 'o:','LineWidth',2.5,'MarkerSize',12,'MarkerEdgeColor','k','MarkerFaceColor','y')%[.49 1 .63]
            plot(nskneePoint(1), nskneePoint(2), 'o:','LineWidth',1.5,'MarkerSize',8,'MarkerEdgeColor','k','MarkerFaceColor','y')
            plot(nsankPoint(1), nsankPoint(2), 'o:','LineWidth',1,'MarkerSize',6,'MarkerEdgeColor','k','MarkerFaceColor','y')             
            ymax = heelPoint(2)-refPoint(2)+1.0 + camera_shift(2);
            ymin = heelPoint(2)-refPoint(2)-.2 + camera_shift(2);
            xmax = heelPoint(1)-refPoint(1)+.5 + camera_shift(1);
            xmin = xmax-(ymax-ymin)*width/height;
            
            %if the biped gets out of frame, shift the bounds
            if nstoePoint(1)<xmin || nstoePoint(2)<ymin
                %these are magic numbers, steplength*# of steps to get out of frame
                camera_shift = -.52663*5 * [cos(slope),sin(slope)];
            end
            
            %Draw the ground, the line intersects with xmin and xmax
            plot([xmin,xmax],[ xmin*tan(slope), xmax*tan(slope)],'Color',[0.396, 0.263, 0.129])
            
            axis([xmin, xmax, ymin, ymax]);
            axis off
            text( xmin ,ymax,strcat('step ' ,num2str(N)) );
            text( xmin ,ymax - .05 ,strcat('t = ' ,num2str(t(i))) );
            drawnow;

            % Add movie frames?
            if ~isempty(movie)
                F = getframe(gcf);
                writeVideo(mov,F);
            end

        end

        if t(i)>=Step_time_sum(N)&&N<stepnum
            N=N+1;
        end

    end

end