function [fstate, xout, tout, out_extra] = walk(xi, steptotal)
% WALK2   Walk the robot for "steptotal" number of steps.
% 
% INPUT:
% + xi = the input state vector, configuration q and their velocites qdot. Passed to
%       step. q = [x, y, heel, ankle, stance knee, hip, swing knee,swing ankle].
% + steptotal = specify the number of steps the walker should take.
% 
% 
% OUTPUT:
% Note: time is the first column in all of these outputs. 
% + fstate: the state after impact on the last step. 
% + xout: all the states beginning from time = 0 to the time 
%           of the last point of contact with the slope.
% + tout: vector of times that match with xout
% + out_extra:  struct of extra outputs,
%     out_extra.t_impacts       %a cell array of n steps. Each cell has a 3x1 array that gives the time when the biped siwtches from: heel->flat,  flat->toe, toe->swing heel
%     out_extra.istate_minus    %an array of n (steps) x 16 doubles. It gives the pre-impact state vector BEFORE heel strike
%     out_extra.istate_plus     %an array of n (steps) x 16 doubles. It gives the post-impact state vector AFTER heel strike
%     out_extra.step_lengths    %an array of n (steps) doubles. It gives the length of each step. 
%     out_extra.clearance_heel  %an array of n (steps) doubles. It gives the minimum heel clearance of each step.
%     out_extra.clearance_toe   %an array of n (steps) doubles. It gives the minimum toe clearance of each step.
%     out_extra.step_time       %an array of n (steps) doubles. It gives the time to complete each step, from heel strike-> next heel strike.


global dim stepnum 
global L la l1 l2 lf 
global stepLengths stepTimeSpans 
global stepclearance_heel stepclearance_toe clearance_heel clearance_toe steplength
global Step_time_sum 
global silent
    if~silent,disp('New Walking Simulation');end

    if size(xi) ~= [1,dim] 
        fprintf('walk error: Input state vector must have %i row and %i cols.\n', size(xi)*[1 0]', size(xi)*[0 1]');
        fstate = []; xout = []; 
        return
    end
  
    %initializion
    x = [];
    t = [];
    stepTimeSpans = [];
    stepLengths=[];
    clearance_heel=[];
    clearance_toe=[];
    abs_heel_clearance = [];
    abs_toe_clearance = [];
    x_postimpacts = zeros(steptotal,dim); % all impact-event states
    x_preimpacts = zeros(steptotal,dim); % all pre-impact-event states
    t_impacts = cell(steptotal,1); % all impact-event times
    tlast = 0;
    
    %run the simulation for n steps
    for stepnum=1:steptotal
        [xafter, tafter, xbefore, tbefore] = step(xi);
        t_impacts{stepnum} = tafter +tlast;
        tafter = tafter(end);
        if~silent,disp(strcat('step ' ,num2str(stepnum), ' , length: ', num2str(steplength)));end

        if isempty(xafter) && isempty(tafter) && isempty(xbefore) && isempty(tbefore)
            fstate = []; xout = []; 
            return
        end

        stepTimeSpans = [stepTimeSpans; tafter];
        t = [t ; tbefore+tlast];
        x = [x ; xbefore];
        
        x_postimpacts(stepnum,:) = xafter;
        x_preimpacts(stepnum,:) = xbefore(end,:);     % assign the last row of xbefore to the "step" row of the x_preimpacts
        xi = xafter;                               % Re-assign initial conditions
        tlast = tafter+tlast;
        stepLengths = [stepLengths; steplength];  % Record step-length based on previous step
        clearance_heel=[clearance_heel;stepclearance_heel(:,1)]; % Calculating heel clearance based on stepclearance_heel   
        clearance_toe=[clearance_toe;stepclearance_toe]; % Calculating toe clearance based on stepclearance_toe
        Step_time_sum=[Step_time_sum;tlast];         % Recording time for one step, tlast;
        
        midswing_i = xbefore(:,6)<0.3 & xbefore(:,6)>-0.35; %this is to avoid minimums near lift off and touch down of the swing leg, which are meaningless
        abs_heel_clearance = [abs_heel_clearance;min(clearance_heel(midswing_i))]; %#ok<*AGROW> %stepclearance_heel(midswing_i,1)];
        abs_toe_clearance = [abs_toe_clearance;min(clearance_toe(midswing_i))];%stepclearance_toe(midswing_i)];
    end
    
    fstate = xafter;
    xout = x;
    tout = t;
    
    out_extra.t_impacts = t_impacts;        %a cell array of n steps. Each cell has a 3x1 array that gives the time when the biped siwtches from: heel->flat,  flat->toe, toe->swing heel
    out_extra.istate_minus = x_preimpacts;  %an array of n (steps) x 16 doubles. It gives the pre-impact state vector BEFORE heel strike
    out_extra.istate_plus = x_postimpacts;  %an array of n (steps) x 16 doubles. It gives the post-impact state vector AFTER heel strike
    out_extra.step_lengths = stepLengths;   %an array of n (steps) doubles. It gives the length of each step. 
    out_extra.clearance_heel = abs_heel_clearance ; %an array of n (steps) doubles. It gives the minimum heel clearance of each step.
    out_extra.clearance_toe = abs_toe_clearance; %an array of n (steps) doubles. It gives the minimum toe clearance of each step.
    out_extra.step_time = stepTimeSpans; %an array of n (steps) doubles. It gives the time to complete each step, from heel strike-> next heel strike.
 
    u_pd = zeros(length(tout),8);
    for i = 1:length(tout)
        u_pd(i,:)= dynamics(tout(i),xout(i,:)','u');

    end
    out_extra.u_pd = u_pd; %an array size(t) x 8 of the control torques trajectory for each joint, in order of the position state vector
end